/*
 Dahdi layer for Quasar E1 interface board driver

 Copyright (C) 2006-2007 Parabel Ltd.
 Author: Mikhail Khelik <mixel.net@gmail.com>

 Copyright (C) 2008-2009 Parabel Ltd.
 Author: Pavel Selivanov <pavel[AT]parabel.ru>

 * Parabel Ltd. reserve the right to distribute project or it's parts
 under the alternative conditions.

 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software Foundation;
 either version 2 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 See the GNU General Public License for more details.

 You should have received a copy of the GNU General Public License along with this program;
 if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/


#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/pci.h>
#include <linux/delay.h>
#include <linux/spinlock.h>
#include <linux/completion.h>
#include <linux/list.h>
#include <linux/types.h>
#include <linux/interrupt.h>

#include "dahdi/kernel.h"

#include "dver.h"
#include "pmc_e1_conf.h"
#include "pmc_e1.h"
#include "e12zap.h"


#ifdef E12ZAP_ONE_SPAN
 #define E1PMC_NUM_SPANS 1                              /* to change you have to rewrite interrupt headler */
#else
 #define E1PMC_NUM_SPANS 8
#endif

struct  e12zap_span
{
    struct dahdi_span span;
    struct dahdi_chan _chans[31];                       /* omit 0 timeslot */
    struct dahdi_chan *chans[31];                       /* omit 0 timeslot */
    int spannumber;		                        /* span number in card */
    int syncprior;			                /* use or not this span for syncronization, and syncsource priority */
    void * pvt;
    int clear_counters;
    unsigned char ec_chunk1[31][DAHDI_CHUNKSIZE];          /* first EC chunk buffer */
    unsigned char ec_chunk2[31][DAHDI_CHUNKSIZE];          /* second EC chunk buffer */
};

struct e12zap_pmc
{
  volatile unsigned int *writechunk;			/* Double-word aligned write memory */
  volatile unsigned int *readchunk;                     /* Double-word aligned read memory */
  dma_addr_t 	readdma;
  dma_addr_t	writedma;
  struct e12zap_span    *e12zap_pmc_span[E1PMC_NUM_SPANS];
  unsigned char num_spans;		                        /* how many spans this card has */
  unsigned char qn_span;                                /*For fast check during the interrupt*/
  char registered;
  void * pvt;
  u32 ovfl_counter;                                     /* keep old value of overflow counter, used in e12zap_delivery */
  char check;                                           /* used for checking interrupt misses */
  int intcount;
  int irqmisses;        //digium defined it as signed...
#ifdef E12ZAP_TASKLET
  int taskletrun;
  int taskletsched;
  char taskletpending;
  int taskletexec;
  int txerrors;
  struct tasklet_struct e12zap_tlet;
#endif
};

struct e12zap_span *qspan_from_dspan (struct dahdi_span *span) {
        return container_of(span, struct e12zap_span, span);
}

void e12zap_find_master (struct e12zap_pmc * pmc_zap)
{
	int i;
	struct e12zap_span *ts;
	struct pmc_e1_controller * tmp_dev = pmc_zap->pvt;
	int syncsrc, master;

	master = tmp_dev->bar0[0x02] & 0x07;
	syncsrc = pmc_zap->e12zap_pmc_span[master]->span.spanno;
	ts = pmc_zap->e12zap_pmc_span[master];

	if (tmp_dev->bar0[0x02] & (MODE1_HOLD|MODE1_FREE))
		syncsrc = 0; //Using internal sync source for CT-BUS

	for (i=0; i<pmc_zap->num_spans; ++i) {
		ts = pmc_zap->e12zap_pmc_span[i];
		ts->span.syncsrc = syncsrc;
	}
}

#ifdef HAVE_SPAN_COUNT
# define DAHDI_COUNT(a,b,c) a.c
#else
# define DAHDI_COUNT(a,b,c) a.b
#endif //HAVE_DAHDI_COUNT

void e12zap_check_alarm (struct e12zap_pmc * pmc_zap)
{
	int i;
        unsigned long reg;
	struct e12zap_span *ts;
	struct pmc_e1_controller * tmp_dev = pmc_zap->pvt;

	for (i=0; i<pmc_zap->num_spans; ++i) {
		ts = pmc_zap->e12zap_pmc_span[i];

		if (framer_checkbit (tmp_dev, i, STAT0_RAIS_HP)) {
			ts->span.alarms |= DAHDI_ALARM_YELLOW;
		} else {
			ts->span.alarms &= ~DAHDI_ALARM_YELLOW;
		}
		if (framer_checkbit (tmp_dev, i, STAT0_LOF_HP)) {
			ts->span.alarms |= DAHDI_ALARM_RED;
			ts->span.alarms &= ~DAHDI_ALARM_YELLOW;
		} else {
			ts->span.alarms &= ~DAHDI_ALARM_RED;
		}
		if (framer_checkbit (tmp_dev, i, STAT0_AIS_HP)) {
			ts->span.alarms |= DAHDI_ALARM_BLUE;
			ts->span.alarms &= ~DAHDI_ALARM_YELLOW;
		} else {
			ts->span.alarms &= ~DAHDI_ALARM_BLUE;
		}
		dahdi_alarm_notify (&ts->span);

		DAHDI_COUNT(ts->span, fascount, count.fas) =
			framer_read_pr (tmp_dev, i, CNT_FAS_HP);

		reg = framer_read_pr (tmp_dev, i, CNT_BPVHI_HP);
                reg <<= 8;
		reg |= framer_read_pr (tmp_dev, i, CNT_BPVLO_HP);
                DAHDI_COUNT(ts->span, bpvcount, count.bpv) = reg;

		reg = framer_read_pr (tmp_dev, i, CNT_CRC4HI_HP);
                reg <<= 8;
		reg |= framer_read_pr (tmp_dev, i, CNT_CRC4LO_HP);
		DAHDI_COUNT(ts->span, crc4count, count.crc4) = reg;

		ts->span.irqmisses = pmc_zap->irqmisses;
	}
}

void e12zap_clearstat (struct e12zap_pmc * pmc_zap) {
	struct pmc_e1_controller * tmp_dev = pmc_zap->pvt;
	struct e12zap_span *ts;
	int i;

	for (i=0; i<pmc_zap->num_spans; ++i) {
		ts = pmc_zap->e12zap_pmc_span[i];

		if (0 == ts->clear_counters) continue; //Already cleared
		--ts->clear_counters;
		if (0 != ts->clear_counters) continue;
		module_dprintk(KERN_DEBUG, "e12zap_clearstat:"
			       "Clearing span %d error counters\n", ts->span.spanno);
		framer_write_pr (tmp_dev, i, CNT_FAS_HP, 0);
		framer_write_pr (tmp_dev, i, CNT_BPVHI_HP, 0);
		framer_write_pr (tmp_dev, i, CNT_BPVLO_HP, 0);
		framer_write_pr (tmp_dev, i, CNT_CRC4HI_HP, 0);
		framer_write_pr (tmp_dev, i, CNT_CRC4LO_HP, 0);
		pmc_zap->irqmisses = 0;
	}
}

static int e12zap_maint(struct dahdi_span *span, int cmd)
{
	//Currently ignore all maint commands
	int res = 0;
	switch(cmd) {
	case DAHDI_MAINT_NONE:
		/* lloop=off, rloop=off */
	case DAHDI_MAINT_LOCALLOOP:
		/* lloop=on, rloop=? */
	case DAHDI_MAINT_REMOTELOOP:
		/* lloop=?, rloop=on */
	case DAHDI_MAINT_LOOPUP:
	case DAHDI_MAINT_LOOPDOWN:
	case DAHDI_MAINT_LOOPSTOP:
                span->mainttimer = span->maintstat = 0;
		module_printk(KERN_WARNING, "e21zap_maint: Use ecfg testing options.\n");
                res = -ENOSYS;
		break;
	default:
		module_printk(KERN_ERR, "e21zap_maint: Unknown maint command: %d\n", cmd);
		res = -ENOTTY;
		break;
	}
	return res;
}

static int e12zap_rbsbits(struct dahdi_chan *chan, int bits)
{
	//trunk for CAS nibles...
	//put CAS nibbles to appropriate position
	struct e12zap_span * e12zap_pmc_span = chan->pvt;
	struct e12zap_pmc * pmc_zap = e12zap_pmc_span->pvt;
	struct pmc_e1_controller * tmp_dev = pmc_zap->pvt;

	module_dprintk(KERN_DEBUG, "e12zap_rbsbits:"
		       "->%x on %s\n", bits, chan->name);

	if (chan->chanpos == 16)
		return 0;

	if(chan->chanpos < 16){
                framer_write (tmp_dev, e12zap_pmc_span->spannumber, 0x5, 0x10 + chan->chanpos, bits);
	}
	else {
                framer_write (tmp_dev, e12zap_pmc_span->spannumber, 0x6, 0x10 + chan->chanpos, bits);
	}
	return 0;
}

static int e12zap_chanconfig(struct dahdi_chan *chan, int sigtype)
{
	//hmm... in E1 case, just nothing to do...
	return 0;
}

static int e12zap_ioctl(struct dahdi_chan *chan, unsigned int cmd, unsigned long data)
{
	switch(cmd) {
	default:
		return -ENOTTY;
	}
}

static int e12zap_close(struct dahdi_chan *chan)
{
	module_dprintk(KERN_DEBUG, "e12zap_close\n");
	module_put(THIS_MODULE);
	return 0;
}

static int e12zap_open(struct dahdi_chan *chan)
{
	module_dprintk(KERN_DEBUG, "e12zap_open\n");
	try_module_get(THIS_MODULE);
	return 0;
}

static int e12zap_shutdown(struct dahdi_span *span)
{
	int wasrunning, i, tmp;
	unsigned char working=0;
	struct e12zap_span *p = qspan_from_dspan(span);
	struct e12zap_pmc * pmc_zap = p->pvt;
	struct pmc_e1_controller * ptmc_dev = pmc_zap->pvt;
	wasrunning = span->flags & DAHDI_FLAG_RUNNING;
	span->flags &= ~DAHDI_FLAG_RUNNING;

	module_dprintk(KERN_DEBUG, "e12zap_shutdown\n");
	if (wasrunning){
		for(i=0;i<pmc_zap->num_spans;i++){
			if(pmc_zap->e12zap_pmc_span[i]->span.flags & DAHDI_FLAG_RUNNING)
				working =1;
		}
	}
	if (!working){
		//disable DMA on recv and transmit
		tmp = ptmc_dev->bar0[0x09];
		tmp &= ~(DMA_WENA|DMA_RENA);
		ptmc_dev->bar0[0x09] = tmp;

		//cancel pending interrupts
		ptmc_dev->bar0[0] = 0;

		//disable DMA interrupt, mask
		tmp = ptmc_dev->bar0[0x01];
		tmp &= ~(0x20000000);
		ptmc_dev->bar0[0x01] = tmp;
	}
	return 0;
}

static int e12zap_startup(struct dahdi_span *span)
{
	int i, tmp, ucTmp1;
	struct e12zap_span *p = qspan_from_dspan(span);
        struct e12zap_pmc *pmc_zap = p->pvt;
	struct pmc_e1_controller * ptmc_dev = pmc_zap->pvt;
       	int alreadyrunning = span->flags & DAHDI_FLAG_RUNNING;

	module_dprintk(KERN_DEBUG, "e12zap_startup: ");

	/* initialize the start value for the entire chunk of last ec buffer */
	for(i = 0; i < span->channels; i++)
	{
		memset(p->ec_chunk1[i],
			DAHDI_LIN2X(0,span->chans[i]),DAHDI_CHUNKSIZE);
		memset(p->ec_chunk2[i],
			DAHDI_LIN2X(0,span->chans[i]),DAHDI_CHUNKSIZE);
	}
	dprintk("", "Using ");

        if(span->lineconfig & DAHDI_CONFIG_HDB3) //set HDB3
	{
                framer_setbit(ptmc_dev, p->spannumber, CFG0_RCV_HDB3_HP);
                framer_setbit(ptmc_dev, p->spannumber, CFG0_XMIT_HDB3_HP);
		dprintk("", "HDB3/");
 	}
        else //set AMI
	{
		framer_clrbit(ptmc_dev, p->spannumber, CFG0_RCV_HDB3_HP);
                framer_clrbit(ptmc_dev, p->spannumber, CFG0_XMIT_HDB3_HP);
		dprintk("", "AMI/");
	}

	if(span->lineconfig & DAHDI_CONFIG_CCS)
	{       //set CSS
                framer_setbit(ptmc_dev, p->spannumber, CFG0_CAS_HP);
		dprintk("", "CCS/");
	}
	else  //set CAS
	{
                framer_clrbit(ptmc_dev, p->spannumber, CFG0_CAS_HP);
		dprintk("", "CAS/");
	}

	/* Basic CRC-4 synchronization processing. */
	framer_clrbit(ptmc_dev, p->spannumber, CFG3_CRC4_DIS_HP);
        /* Automatic CRC-interworking */
	framer_clrbit(ptmc_dev, p->spannumber, CFG3_CRC4_MANUAL_HP);

	/* Should I strict DAHDI's readme. See quasar/crc4.txt notes. */
	if (span->lineconfig & DAHDI_CONFIG_CRC4) {
		dprintk("", "CRC4 ");
	} else {
 		dprintk("", "non-CRC4 ");
	}

        dprintk("", "on span %s, ", span->name);
	if (p->syncprior) //if !0 allow to use for syncronisation.
        {
	    tmp = ptmc_dev->bar0[0x06]; //get bs/ls
	    ucTmp1 = ~(1<<p->spannumber);
	    tmp&= ucTmp1;
	    ptmc_dev->bar0[0x06] = tmp; //restore bs/ls
            dprintk("", "can be the sync. source\n");
	}
	else
	{
	    tmp = ptmc_dev->bar0[0x06]; //get bs/ls
	    ucTmp1 = (1<<p->spannumber);
	    tmp|= ucTmp1;
	    ptmc_dev->bar0[0x06] = tmp; //restore bs/ls
            dprintk("", "can not be the sync. source\n");
	}

	if (!alreadyrunning) {
		module_dprintk(KERN_DEBUG, "Enable DMA...\n");
	// Only if we're not already going
	//allow DMA on recv and transmit
		tmp = ptmc_dev->bar0[0x09];
		tmp |= (DMA_WENA | DMA_RENA);
		ptmc_dev->bar0[0x09] = tmp;
		//allow DMA interrupts, unmask, allow CASsig interrupt
		tmp = ptmc_dev->bar0[0x01];
		tmp |= 0x20000000;
		ptmc_dev->bar0[0x01] = tmp;
		span->flags |= DAHDI_FLAG_RUNNING;
	}
	p->clear_counters = 3;
	return 0;
}

static int e12zap_spanconfig(struct dahdi_span *span, struct dahdi_lineconfig *lc)
{
	struct e12zap_span * p = qspan_from_dspan(span);;

	module_dprintk(KERN_DEBUG, "e12zap_spanconfig:"
		       "Configuring span %d, span=%p\n", span->spanno, span);
	span->lineconfig = lc->lineconfig;
	span->txlevel = lc->lbo;
	span->rxlevel = 0;
	p->syncprior = lc->sync; 	//lineconfig.sync contain the priority of the sysnc source 1-highest 1, 0 - not use this span for the sync at all.
	/* If we're already running, then go ahead and apply the changes */
	if (span->flags & DAHDI_FLAG_RUNNING)
	{
		module_printk(KERN_INFO, "Starting span!\n");
		return e12zap_startup(span);
	}
	module_dprintk(KERN_DEBUG, "Done with spanconfig!\n");
	return 0;
}

void e12zap_proc(struct e12zap_pmc * pmc_zap);
#ifdef E12ZAP_TASKLET
static void e12zap_tasklet(unsigned long data)
{
	struct e12zap_pmc * pmc_zap = (struct e12zap_pmc *)data;
	pmc_zap->taskletrun++;
	if (pmc_zap->taskletpending) {
		pmc_zap->taskletexec++;
		e12zap_proc(pmc_zap);
	}
	pmc_zap->taskletpending = 0;
}
#endif

int e12zap_dacs(struct dahdi_chan *dst, struct dahdi_chan *src);
#ifdef HAVE_SPAN_OPS
static const struct dahdi_span_ops e12zap_span_ops = {
	.owner = THIS_MODULE,
	.spanconfig = e12zap_spanconfig,
	.chanconfig = e12zap_chanconfig,
	.startup = e12zap_startup,
	.shutdown = e12zap_shutdown,
	.rbsbits = e12zap_rbsbits,
	.maint = e12zap_maint,
	.open = e12zap_open,
	.close = e12zap_close,
	.ioctl = e12zap_ioctl,
	.dacs = e12zap_dacs,
};
void e12zap_setops (struct dahdi_span *span) {
        span->ops = &e12zap_span_ops;
}
#else //HAVE_SPAN_OPS
void e12zap_setops (struct dahdi_span *span) {
	    span->spanconfig = e12zap_spanconfig;
	    span->chanconfig = e12zap_chanconfig;
	    span->startup = e12zap_startup;
	    span->shutdown = e12zap_shutdown;
	    span->rbsbits = e12zap_rbsbits;
	    span->maint = e12zap_maint;
	    span->open = e12zap_open;
	    span->close  = e12zap_close;
	    span->ioctl = e12zap_ioctl;
	    span->dacs = e12zap_dacs;
#ifdef HAVE_SPAN_OWNER
            span->owner = THIS_MODULE;
#endif //HAVE_SPAN_OWNER
}
#endif //HAVE_SPAN_OPS

int e12zap_register(struct pmc_e1_controller * ptmc_dev)
{
    struct pci_dev *pdev = ptmc_dev->pci_device;
    int tmp, j, x;
    struct e12zap_pmc * pmc_zap;
    ptmc_dev->priv_data = kmalloc(sizeof(struct e12zap_pmc), GFP_KERNEL);
    if (ptmc_dev->priv_data == NULL) return -ENOMEM;
    memset(ptmc_dev->priv_data,0x0,sizeof(struct e12zap_pmc));
    pmc_zap = (struct e12zap_pmc *) ptmc_dev->priv_data;
    pmc_zap->pvt = ptmc_dev;
    pmc_zap->writechunk  = (unsigned int *)pci_alloc_consistent(pdev, DAHDI_CHUNKSIZE * 1024, &pmc_zap->writedma); //buffers for 16 frames on input and output;
    if (pmc_zap->writechunk == NULL)
    {
	    module_printk(KERN_INFO, "Unable to allocate DMA buffer\n");
	    kfree(ptmc_dev->priv_data);
	    return -ENOMEM;
    }
    //  printk("DMA phys address=%x\n", pmc_zap->writedma);

    pmc_zap->readchunk = pmc_zap->writechunk + (DAHDI_CHUNKSIZE*128); //double-word address;
    pmc_zap->readdma = pmc_zap->writedma + (DAHDI_CHUNKSIZE * 512); //byte address

#ifdef E12ZAP_ONE_SPAN
    pmc_zap->num_spans = E1PMC_NUM_SPANS;
#else
    pmc_zap->num_spans = ptmc_dev->e1_amount + 1;
    if (pmc_zap->num_spans == 2)
	pmc_zap->qn_span = 2;
    else
        pmc_zap->qn_span = 4;
#endif
    /* Initialize Write/Buffers to all blank data */
    memset((void *)pmc_zap->writechunk,0x00, DAHDI_CHUNKSIZE * 512);
    memset((void *)pmc_zap->readchunk,0xff, DAHDI_CHUNKSIZE * 512);

    pci_set_master(pdev);

    //allocate memory for the spans
    for (tmp = 0; tmp < pmc_zap->num_spans; tmp++)
    {
	    pmc_zap->e12zap_pmc_span[tmp] = kmalloc(sizeof(struct e12zap_span), GFP_KERNEL);
	    if (pmc_zap->e12zap_pmc_span[tmp])
	    {
		    memset(pmc_zap->e12zap_pmc_span[tmp], 0, sizeof(struct e12zap_span));
	    }
	    else
	    {   //free already allocated
		    for (j=(tmp-1);j >=0; j--)
		    {
			    kfree(pmc_zap->e12zap_pmc_span[tmp]);
		    }
		    goto unreg_all;
	    }
    }
    //initialize spans
    for (tmp = 0; tmp < pmc_zap->num_spans; tmp++)
    {
	    pmc_zap->e12zap_pmc_span[tmp]->spannumber=tmp;
	    pmc_zap->e12zap_pmc_span[tmp]->pvt = pmc_zap;
	    pmc_zap->e12zap_pmc_span[tmp]->span.channels = 31;
	    pmc_zap->e12zap_pmc_span[tmp]->span.spantype = "E1";
	    pmc_zap->e12zap_pmc_span[tmp]->span.manufacturer = "Parabel ltd.";
	    sprintf(pmc_zap->e12zap_pmc_span[tmp]->span.name,
		    "Quasar_%d_%d", ptmc_dev->dev_number,tmp+1);
	    sprintf(pmc_zap->e12zap_pmc_span[tmp]->span.desc,
		    "Parabel Quasar %d Span %d",ptmc_dev->dev_number, tmp+1);
	    snprintf(pmc_zap->e12zap_pmc_span[tmp]->span.location,
		     sizeof(pmc_zap->e12zap_pmc_span[tmp]->span.location) - 1,
		     "PCI Bus %02d Slot %02d", pdev->bus->number,
		     PCI_SLOT(pdev->devfn) + 1);
	    snprintf(pmc_zap->e12zap_pmc_span[tmp]->span.devicetype,
		     sizeof(pmc_zap->e12zap_pmc_span[tmp]->span.devicetype) - 1,
		     "Quasar-%d PCI/PCIe/PMC, hw:0x%X",
		     pmc_zap->num_spans, ptmc_dev->bar0[0x0D]);
	    pmc_zap->e12zap_pmc_span[tmp]->span.chans = pmc_zap->e12zap_pmc_span[tmp]->chans;
	    pmc_zap->e12zap_pmc_span[tmp]->span.flags = DAHDI_FLAG_RBS;
	    pmc_zap->e12zap_pmc_span[tmp]->span.linecompat = DAHDI_CONFIG_HDB3 | DAHDI_CONFIG_CCS | DAHDI_CONFIG_CRC4 | DAHDI_CONFIG_AMI;  //which modes I can work
	    pmc_zap->e12zap_pmc_span[tmp]->span.deflaw = DAHDI_LAW_ALAW;

	    e12zap_setops (&pmc_zap->e12zap_pmc_span[tmp]->span);

	    init_waitqueue_head(&(pmc_zap->e12zap_pmc_span[tmp]->span.maintq));

	    for (x=0; x < pmc_zap->e12zap_pmc_span[tmp]->span.channels; x++)
	    {
                    pmc_zap->e12zap_pmc_span[tmp]->chans[x] = &pmc_zap->e12zap_pmc_span[tmp]->_chans[x];
		    sprintf(pmc_zap->e12zap_pmc_span[tmp]->chans[x]->name, "Quasar_%d_%d_%d", ptmc_dev->dev_number, tmp + 1, x + 1);
		    pmc_zap->e12zap_pmc_span[tmp]->chans[x]->sigcap = DAHDI_SIG_EM | DAHDI_SIG_CLEAR | DAHDI_SIG_EM_E1 |
			    DAHDI_SIG_FXSLS | DAHDI_SIG_FXSGS |
			    DAHDI_SIG_FXSKS | DAHDI_SIG_FXOLS |
			    DAHDI_SIG_FXOGS | DAHDI_SIG_FXOKS |
			    DAHDI_SIG_CAS | DAHDI_SIG_SF | DAHDI_SIG_DACS_RBS;
		    pmc_zap->e12zap_pmc_span[tmp]->chans[x]->pvt = pmc_zap->e12zap_pmc_span[tmp];
		    pmc_zap->e12zap_pmc_span[tmp]->chans[x]->chanpos = x + 1; //timeslot number
	    }
	    //attach to dahdi
	    if (dahdi_register(&pmc_zap->e12zap_pmc_span[tmp]->span, 0))
	    {
		    module_printk(KERN_ERR, "Unable to register span Quasar_%d_%d with dahdi!\n", ptmc_dev->dev_number, tmp+1);
		    for (j=(tmp-1);j >=0; j--)
		    {
			    module_printk(KERN_INFO, "Unregister span %d\n", j);
			    dahdi_unregister(&pmc_zap->e12zap_pmc_span[tmp]->span);
		    }
		    //free spans
		    for (j = 0; j < pmc_zap->num_spans; j++)
		    {
			    kfree(pmc_zap->e12zap_pmc_span[j]);
		    }
		    goto unreg_all;
	    }
    }
#ifdef E12ZAP_TASKLET
    tasklet_init(&pmc_zap->e12zap_tlet, e12zap_tasklet, (unsigned long)pmc_zap);
#endif
    //initialize hardware
    tmp = pmc_zap->readdma;
    ptmc_dev->bar0[0x0a] = tmp;
    tmp = pmc_zap->writedma;
    ptmc_dev->bar0[0x0c] = tmp;
    tmp = DMA_IFQ0 | DMA_SIZE0;
    ptmc_dev->bar0[0x09] = tmp;

    pmc_zap->registered = 1;
    return 0;
unreg_all:
	module_printk(KERN_ERR, "e12zap_register: Not enough free resources!\n");
	pci_free_consistent(pdev, DAHDI_CHUNKSIZE * 1024, (void *) pmc_zap->writechunk, pmc_zap->writedma);
	kfree(ptmc_dev->priv_data);
	return -ENOMEM;
}
int e12zap_unregister(struct pmc_e1_controller * ptmc_dev)
{
	struct pci_dev *pdev = ptmc_dev->pci_device;
	struct e12zap_pmc * pmc_zap;
	int j, tmp;
	pmc_zap = (struct e12zap_pmc *) ptmc_dev->priv_data;
	if(pmc_zap)
	{
		if (pmc_zap->registered)
		{
			//detach from dahdi
			for(j=0;j<pmc_zap->num_spans;j++)
			{
				dahdi_unregister(&pmc_zap->e12zap_pmc_span[j]->span);
				kfree(pmc_zap->e12zap_pmc_span[j]);
			}
			pci_free_consistent(pdev, DAHDI_CHUNKSIZE * 1024, (void *) pmc_zap->writechunk, pmc_zap->writedma);
			kfree(ptmc_dev->priv_data);
			//deinitialize hardware

			//disable DMA on recv and transmit
			tmp = ptmc_dev->bar0[0x09];
			tmp &= ~(DMA_WENA|DMA_RENA);
			ptmc_dev->bar0[0x09] = tmp;

			//disable DMA interrupt, mask
			tmp = ptmc_dev->bar0[0x01];
			tmp &= ~(0x02000000);
			ptmc_dev->bar0[0x01] = tmp;
		}
	}
    return 0;
}

void e12zap_proc(struct e12zap_pmc * pmc_zap)
{
    int j1,i1;
    for(j1=0; j1 < pmc_zap->num_spans; j1++)
    {
	if( pmc_zap->e12zap_pmc_span[j1]->span.flags &  DAHDI_FLAG_RUNNING)
	{
	    /*Echo cancelation*/
		for(i1=0; i1 < pmc_zap->e12zap_pmc_span[j1]->span.channels; i1++)
		{
			dahdi_ec_chunk(pmc_zap->e12zap_pmc_span[j1]->span.chans[i1],
				    pmc_zap->e12zap_pmc_span[j1]->span.chans[i1]->readchunk,
				    pmc_zap->e12zap_pmc_span[j1]->ec_chunk2[i1]);
			memcpy(pmc_zap->e12zap_pmc_span[j1]->ec_chunk2[i1], pmc_zap->e12zap_pmc_span[j1]->ec_chunk1[i1],
			       DAHDI_CHUNKSIZE);
			memcpy(pmc_zap->e12zap_pmc_span[j1]->ec_chunk1[i1], pmc_zap->e12zap_pmc_span[j1]->span.chans[i1]->writechunk,
			       DAHDI_CHUNKSIZE);
		}
		dahdi_receive(&pmc_zap->e12zap_pmc_span[j1]->span);
	}
    }
    for(j1=0; j1 < pmc_zap->num_spans; j1++)
    {
	    if( pmc_zap->e12zap_pmc_span[j1]->span.flags &  DAHDI_FLAG_RUNNING)
	    {
		    dahdi_transmit(&pmc_zap->e12zap_pmc_span[j1]->span);
	    }
    }
}

int e12zap_delivery(struct pmc_e1_controller * ptmc_dev)
{
    struct e12zap_pmc * pmc_zap = ptmc_dev->priv_data;
    unsigned int keep,i1, n1, j1,tmp1,tmp2, shift;
    /* 1. Free and fill system DMA buffers prepare to call dahdi_receive and dahdi_transmit.
	 2. Check and renew CAS signalization.
	 3. Check loosing of sync.
	 4. Perform echo cancelation on RX chanks.
	 5. Call dahdi_receive and dahdi_transmit. //process independently
	 */
	//check in which part of DMA buffer we are...

    if(!ptmc_dev->priv_data)
    {
	    return -1;
    }
#ifdef E12ZAP_TASKLET
    if (!pmc_zap->taskletpending) {
	    pmc_zap->taskletpending = 1;
	    pmc_zap->taskletsched++;
	    tasklet_schedule(&pmc_zap->e12zap_tlet);
    } else {
	    module_dprintk(KERN_DEBUG, "Miss tasklet. Shed:%d Exec:%d Run:%d, TX_Err: %d\n", pmc_zap->taskletsched, pmc_zap->taskletexec, pmc_zap->taskletrun, pmc_zap->txerrors++);
	    pmc_zap->txerrors++;
    }
#endif
	//check DMA overflow counter;
    	keep = (ptmc_dev->bar0[0x09] & 0xff000000) >> 24;
    if (keep != pmc_zap->ovfl_counter)
	module_dprintk(KERN_DEBUG, "DMA OVFL:%d\n", pmc_zap->ovfl_counter);
	pmc_zap->ovfl_counter = keep;

	keep = ptmc_dev->bar0[0x0b]; //here we get relative addres of current dma operating position.

#ifdef E12ZAP_ONE_SPAN
	if (keep < (DAHDI_CHUNKSIZE * 256)) //we are on the first part of buffer, serve second part of the buffer;
	{
            if (pmc_zap->check == 1) {
                ++pmc_zap->irqmisses;
                module_dprintk(KERN_DEBUG, "Miss chunk in 1st part.\n");
            }
	    pmc_zap->check = 1;
//	    if (keep > 0x80) printk("e12zap: Possible data loss=0x%x\n", keep);
	    for (i1 = 0; i1 < DAHDI_CHUNKSIZE; i1++)
	    {
		for(j1=0;j1<16;j1++)      //chan
		{
		    shift = j1 +  64 * i1 + DAHDI_CHUNKSIZE * 64;
		    tmp2 = pmc_zap->readchunk[shift];
		    tmp1 = 0;
		    if (j1==0)  //for 1 timeslot
		    {

			pmc_zap->e12zap_pmc_span[0]->span.chans[0]->readchunk[i1] = 0xff & tmp2;
			//transmit data from dahdi 1ts;
			tmp1 |=  pmc_zap->e12zap_pmc_span[0]->span.chans[0]->writechunk[i1];

		    }
		    else
		    {
			pmc_zap->e12zap_pmc_span[0]->span.chans[j1*2-1]->readchunk[i1] = 0xff & (tmp2 >> 8);
			pmc_zap->e12zap_pmc_span[0]->span.chans[j1*2]->readchunk[i1] = 0xff & tmp2;
			//transmit data from dahdi 2-31 ts;
			tmp1 |= (pmc_zap->e12zap_pmc_span[0]->span.chans[j1*2-1]->writechunk[i1] << 8);
			tmp1 |= (pmc_zap->e12zap_pmc_span[0]->span.chans[j1*2]->writechunk[i1]);
		    }
		    //			printk("write = %x\n",tmp1);
			pmc_zap->writechunk[shift] = tmp1;
		}

	    }

	}
	else //we are in second part of the buffer, serve first part of the buffer
	{
            if (pmc_zap->check == 2) {
                ++pmc_zap->irqmisses;
                module_dprintk(KERN_DEBUG, "Miss chunk in 2nd part.\n");
            }
	    pmc_zap->check = 2;
//	    if (keep > 0x880) printk("e12zap: Possible data loss=0x%x\n", keep);
	    for (i1 = 0; i1 < DAHDI_CHUNKSIZE; i1++)
	    {
		for(j1=0;j1<16;j1++)      //chan
		{
		    shift = j1 +  64 * i1;
		    tmp2 = pmc_zap->readchunk[shift];
		    tmp1 = 0;
		    if (j1==0)  //for 1 timeslot
		    {
			pmc_zap->e12zap_pmc_span[0]->span.chans[0]->readchunk[i1] = 0xff & tmp2;
			//transmit data from dahdi 1ts;
			tmp1 |=  pmc_zap->e12zap_pmc_span[0]->span.chans[0]->writechunk[i1];

		    }
		    else
		    {
			pmc_zap->e12zap_pmc_span[0]->span.chans[j1*2-1]->readchunk[i1] = 0xff & (tmp2 >> 8);
			pmc_zap->e12zap_pmc_span[0]->span.chans[j1*2]->readchunk[i1] = 0xff & tmp2;

			//transmit data from dahdi 2-31 ts;
			tmp1 |= (pmc_zap->e12zap_pmc_span[0]->span.chans[j1*2-1]->writechunk[i1] << 8);
			tmp1 |= (pmc_zap->e12zap_pmc_span[0]->span.chans[j1*2]->writechunk[i1]);
		    }
		    pmc_zap->writechunk[shift] = tmp1;
		}
	    }
	}
#else //E12ZAP_ONE_SPAN
	if (keep < (DAHDI_CHUNKSIZE * 256)) //we are on the first part of buffer, serve second part of the buffer;
	{
            if (pmc_zap->check == 1) {
                ++pmc_zap->irqmisses;
                module_dprintk(KERN_DEBUG, "Miss chunk in 1st part.\n");
            }
	    pmc_zap->check = 1;
//	    if (keep > 0x80) printk("e12zap: Possible data loss=0x%x\n", keep);
	    for (i1 = 0; i1 < DAHDI_CHUNKSIZE; i1++)
	    {
		for(n1=0; n1 < pmc_zap->qn_span; n1 ++)      //span
		{
		    for(j1=0;j1<16;j1++)      //chan
		    {
			shift = j1 + 16 * n1 +  64 * i1 + DAHDI_CHUNKSIZE * 64;
			tmp2 = pmc_zap->readchunk[shift];
			tmp1 = 0;
			if (j1==0)  //for 1 timeslot
			{
			    //transmit data from dahdi 1ts;
			    if (pmc_zap->num_spans > 4)
			    {
				pmc_zap->e12zap_pmc_span[4+n1]->span.chans[0]->readchunk[i1] = 0xff & (tmp2 >> 16);
				tmp1 =  (pmc_zap->e12zap_pmc_span[4+n1]->span.chans[0]->writechunk[i1] << 16);
			    }
			    pmc_zap->e12zap_pmc_span[n1]->span.chans[0]->readchunk[i1] = 0xff & tmp2;
			    tmp1 |=  pmc_zap->e12zap_pmc_span[n1]->span.chans[0]->writechunk[i1];

			}
			else
			{
			    //transmit data from dahdi 2-31 ts;
			    if (pmc_zap->num_spans > 4)
			    {
				pmc_zap->e12zap_pmc_span[4+n1]->span.chans[j1*2-1]->readchunk[i1] = tmp2 >> 24;
				pmc_zap->e12zap_pmc_span[4+n1]->span.chans[j1*2]->readchunk[i1] = 0xff & (tmp2 >> 16);
				tmp1 =  (pmc_zap->e12zap_pmc_span[4+n1]->span.chans[j1*2-1]->writechunk[i1] << 24);
				tmp1 |= (pmc_zap->e12zap_pmc_span[4+n1]->span.chans[j1*2]->writechunk[i1] << 16);
			 
			    }
			    pmc_zap->e12zap_pmc_span[n1]->span.chans[j1*2-1]->readchunk[i1] = 0xff & (tmp2 >> 8);
			    pmc_zap->e12zap_pmc_span[n1]->span.chans[j1*2]->readchunk[i1] = 0xff & tmp2;
			    tmp1 |= (pmc_zap->e12zap_pmc_span[n1]->span.chans[j1*2-1]->writechunk[i1] << 8);
			    tmp1 |= (pmc_zap->e12zap_pmc_span[n1]->span.chans[j1*2]->writechunk[i1]);
			}
			//			printk("write = %x\n",tmp1);
			pmc_zap->writechunk[shift] = tmp1;
		    }
		}
	    }

	}
	else //we are in second part of the buffer, serve first part of the buffer
	{
            if (pmc_zap->check == 2) {
                ++pmc_zap->irqmisses;
                module_dprintk(KERN_DEBUG, "Miss chunk in 2nd part.\n");
            }
	    pmc_zap->check = 2;
//	    if (keep > 0x880) printk("e12zap: Possible data loss=0x%x\n", keep);
	    for (i1 = 0; i1 < DAHDI_CHUNKSIZE; i1++)
		{
		    for(n1=0; n1 < pmc_zap->qn_span; n1 ++)      //span
		    {
			for(j1=0;j1<16;j1++)      //chan
			{
			    shift = j1 + 16 * n1 +  64 * i1;
			    tmp2 = pmc_zap->readchunk[shift];
			    tmp1 = 0;
			    if (j1==0)  //for 1 timeslot
			    {
				//transmit data from dahdi 1ts;
				if (pmc_zap->num_spans > 4)
				{
				    pmc_zap->e12zap_pmc_span[4+n1]->span.chans[0]->readchunk[i1] = 0xff & (tmp2 >> 16);
				    tmp1 =  (pmc_zap->e12zap_pmc_span[4+n1]->span.chans[0]->writechunk[i1] << 16);
				}
				pmc_zap->e12zap_pmc_span[n1]->span.chans[0]->readchunk[i1] = 0xff & tmp2;
				tmp1 |=  pmc_zap->e12zap_pmc_span[n1]->span.chans[0]->writechunk[i1];

			    }
			    else
			    {
				//transmit data from dahdi 2-31 ts;
				if (pmc_zap->num_spans > 4)
				{
				    pmc_zap->e12zap_pmc_span[4+n1]->span.chans[j1*2-1]->readchunk[i1] = tmp2 >> 24;
				    pmc_zap->e12zap_pmc_span[4+n1]->span.chans[j1*2]->readchunk[i1] = 0xff & (tmp2 >> 16);

				    tmp1 =  (pmc_zap->e12zap_pmc_span[4+n1]->span.chans[j1*2-1]->writechunk[i1] << 24);
				    tmp1 |= (pmc_zap->e12zap_pmc_span[4+n1]->span.chans[j1*2]->writechunk[i1] << 16);
				}
				pmc_zap->e12zap_pmc_span[n1]->span.chans[j1*2-1]->readchunk[i1] = 0xff & (tmp2 >> 8);
				pmc_zap->e12zap_pmc_span[n1]->span.chans[j1*2]->readchunk[i1] = 0xff & tmp2;

			        tmp1 |= (pmc_zap->e12zap_pmc_span[n1]->span.chans[j1*2-1]->writechunk[i1] << 8);
				tmp1 |= (pmc_zap->e12zap_pmc_span[n1]->span.chans[j1*2]->writechunk[i1]);
			    }
			    pmc_zap->writechunk[shift] = tmp1;
			}
		    }
		}
	}
#endif
	//Check signalisation
	for(i1=0;i1 < pmc_zap->num_spans; i1++) //choose span, i1 - span number;
	{
		
		if(pmc_zap->e12zap_pmc_span[i1]->span.flags & DAHDI_FLAG_RUNNING)
		{
			//read interrupt status
			if (framer_read(ptmc_dev, i1, 0x4, 0x1e) & 0x01)
			{
				//printk("CAS sig changed!\n");
				for(j1=1;j1<16;j1++) //chans < 16                            dahdi channels [1:15]  - chans [0:14]
				{
					tmp1 = framer_read(ptmc_dev, i1, 0x9, 0x10 + j1) & 0x0f;
					if (tmp1 != pmc_zap->e12zap_pmc_span[i1]->span.chans[j1-1]->rxsig)
					{
						if(!(pmc_zap->e12zap_pmc_span[i1]->span.chans[j1-1]->sig & DAHDI_SIG_CLEAR))
						{
						  //      printk("Call, tmp1 =%x on ts:%d\n", tmp1, j1);
							dahdi_rbsbits(pmc_zap->e12zap_pmc_span[i1]->span.chans[j1-1], tmp1);
						}
					}
				}
				for(j1=1;j1<16;j1++) //chans > 16                            dahdi channels [17:31] -chans [16:30]
				{
					tmp1 = framer_read(ptmc_dev, i1, 0xa, 0x10 + j1) & 0x0f;
					if (tmp1 != pmc_zap->e12zap_pmc_span[i1]->span.chans[15+j1]->rxsig)
					{
						if(!(pmc_zap->e12zap_pmc_span[i1]->span.chans[15+j1]->sig & DAHDI_SIG_CLEAR))
						{
						  //      printk("Call, tmp1 =%x on ts:%d\n", tmp1, j1+15);
							dahdi_rbsbits(pmc_zap->e12zap_pmc_span[i1]->span.chans[15+j1], tmp1);
						}
					}
				}
			}
		}
	}

/*This section can be used to pass fake CAS using PTMC_E1_PASS_ABCD_ZAP_FAKE ioctl.
	//Put fake signalisations
#define GET_CHAN_(a) ((a >> 4) & 0xFFFFF)
#define GET_SPAN_(a) (a >> 9)
	int_sig = ptmc_dev->test_sig;
	if ((GET_SPAN_(int_sig) < pmc_zap->num_spans) && (GET_CHAN_(int_sig) < 30))
	{
	    if ((int_sig & 0xf) != pmc_zap->e12zap_pmc_span[GET_SPAN_(int_sig)]->span.chans[GET_CHAN_(int_sig)].rxsig)
	    {
		if(!(pmc_zap->e12zap_pmc_span[GET_SPAN_(int_sig)]->span.chans[GET_CHAN_(int_sig)].sig & DAHDI_SIG_CLEAR))
		{
		    printk("e12zap: Passing sig: span: %d channel:%d, abcd=%01x\n", GET_SPAN_(int_sig)+1, GET_CHAN_(int_sig)+1, (int_sig & 0xf));
		    dahdi_rbsbits(&pmc_zap->e12zap_pmc_span[GET_SPAN_(int_sig)]->span.chans[GET_CHAN_(int_sig)], int_sig & 0xf);
		}
	    }
	 }
*/
#ifndef E12ZAP_TASKLET
	e12zap_proc(pmc_zap);
#endif
	++pmc_zap->intcount;;
        if (0 == (pmc_zap->intcount & 127)) {
	    e12zap_check_alarm (pmc_zap);
	    e12zap_find_master (pmc_zap);
            e12zap_clearstat (pmc_zap);
	}

	return 0;
}

int e12zap_dacs(struct dahdi_chan *dst, struct dahdi_chan *src)
{
	struct e12zap_span * e12zap_pmc_span_dst = dst->pvt;
	struct e12zap_pmc * pmc_zap = e12zap_pmc_span_dst->pvt;
	struct pmc_e1_controller * tmp_dev = pmc_zap->pvt;

	if (!dst) return -1;
	module_dprintk(KERN_DEBUG, "e12zap_dacs:\n");

	if(src)
	{
		struct e12zap_span * e12zap_pmc_span_src = src->pvt;
		if(e12zap_pmc_span_dst->pvt != e12zap_pmc_span_src->pvt) //trying to cross-connect timeslots from different boards
		{
			module_dprintk(KERN_DEBUG, "Can't switch timeslots from different boards, using soft switch\n");
			//dst
			if (e12zap_pmc_span_dst->spannumber < 4)
			{
				tmp_dev->bar0[0x1000 + 0x100 * e12zap_pmc_span_dst->spannumber + dst->chanpos] = CTB_TO_E1_TABLE(4, (dst->chanpos + 32 * e12zap_pmc_span_dst->spannumber));
			}
			else
			{
				tmp_dev->bar0[0x1000 + 0x100 * e12zap_pmc_span_dst->spannumber + dst->chanpos] = CTB_TO_E1_TABLE(5, (dst->chanpos + 32 * (e12zap_pmc_span_dst->spannumber - 4)));
			}
			return -1;
		}
		module_dprintk(KERN_DEBUG, "Switching %d:%d(in)/%d:%d(out).\n", e12zap_pmc_span_dst->spannumber, dst->chanpos, e12zap_pmc_span_src->spannumber, src->chanpos);
                //4 -> 6
                //5 -> 7
		//cross-connect
		if (e12zap_pmc_span_src->spannumber < 4)
		{
			//tmp_dev->bar0[0x1000 + 0x100 * e12zap_pmc_span_dst->spannumber + dst->chanpos] = CTB_TO_E1_TABLE(4, (src->chanpos + 32 * e12zap_pmc_span_src->spannumber));
			tmp_dev->bar0[0x1000 + 0x100 * e12zap_pmc_span_dst->spannumber + dst->chanpos] = CTB_TO_E1_TABLE(6, (src->chanpos + 32 * e12zap_pmc_span_src->spannumber));
		}
		else
		{
			tmp_dev->bar0[0x1000 + 0x100 * e12zap_pmc_span_dst->spannumber + dst->chanpos] = CTB_TO_E1_TABLE(7, (src->chanpos + 32 * (e12zap_pmc_span_src->spannumber - 4)));
		}
	}
	else
	{
	    //return to default crossing
		module_dprintk(KERN_DEBUG, "Return to default crossing on %d:%d\n", e12zap_pmc_span_dst->spannumber, dst->chanpos);

		if (e12zap_pmc_span_dst->spannumber < 4)
		{
			tmp_dev->bar0[0x1000 + 0x100 * e12zap_pmc_span_dst->spannumber + dst->chanpos] = CTB_TO_E1_TABLE(4, (dst->chanpos + 32 * e12zap_pmc_span_dst->spannumber));
		}
		else
		{
			tmp_dev->bar0[0x1000 + 0x100 * e12zap_pmc_span_dst->spannumber + dst->chanpos] = CTB_TO_E1_TABLE(5, (dst->chanpos + 32 * (e12zap_pmc_span_dst->spannumber - 4)));
		}
	}
	return 0;
}
